#include <Arduino_LSM6DS3.h>
#include <ArduinoBLE.h>

#define BLE_DEVICE_NAME "Breath Rate Detector"
#define BLE_LOCAL_NAME "Breath Rate Detector"
const char* UUID_Service = "19b10010-e8f2-537e-4f6c-d104768a1214";
const char* UUID_ax = "19b10011-e8f2-537e-4f6c-d104768a1214";
const char* UUID_ay = "19b10012-e8f2-537e-4f6c-d104768a1214";
const char* UUID_az = "19b10013-e8f2-537e-4f6c-d104768a1214";
const char* UUID_gx = "19b10014-e8f2-537e-4f6c-d104768a1214";
const char* UUID_gy = "19b10015-e8f2-537e-4f6c-d104768a1214";
const char* UUID_gz = "19b10016-e8f2-537e-4f6c-d104768a1214";

// Define a BLE service and characteristic
BLEService Service(UUID_Service); // Custom service UUID
BLEFloatCharacteristic chaAx(UUID_ax, BLERead | BLENotify); 
BLEFloatCharacteristic chaAy(UUID_ay, BLERead | BLENotify); 
BLEFloatCharacteristic chaAz(UUID_az, BLERead | BLENotify); 
BLEFloatCharacteristic chaGx(UUID_gx, BLERead | BLENotify); 
BLEFloatCharacteristic chaGy(UUID_gy, BLERead | BLENotify); 
BLEFloatCharacteristic chaGz(UUID_gz, BLERead | BLENotify); 


const int sampling = 1000; // Sampling period in milliseconds (1 Hz)
const int window = 5000;  // Sliding window size in ms (5 seconds)
float acc_Buffer[1000];        // Buffer for acceleration
int bufferIndex_a = 0;
float threshold = 1.02; // Threshold to detect significant movement (adjust based on your needs)

// Filter coefficients (from matlab script: 0.1-0.49 Hz)

float b[21] =  {0.00931683037284933, 0,	-0.0931683037284933,	0,	0.419257366778220,	0,	-1.11801964474192,	0,	1.95653437829836,	0,	-2.34784125395803,	0,	1.95653437829836,	0,	-1.11801964474192,	0,	0.419257366778220,	0,	-0.0931683037284933,	0,	0.00931683037284933};
float a[21] = {1,	7.70916073863389,	26.2193801260813,	52.9090893787506,	74.1857165778666,	83.2831166956718,	82.1591744709078,	69.9143748679608,	48.9193599755739,	29.4495226454652,	16.5180734695797,	7.91051329362060,	2.83326241996057,	0.956145139883915,	0.374057155228569,	0.0703549907427079,	-0.00271358318385719,	0.00614164964651132,	0.00136790282939128,	-0.000653542477654263,	8.73430659265187e-05}; 

// Filtered values (state variables)
float x[21] = {0};  // Input history
float y[21] = {0};  // Output history

    float ax, ay, az;		// x, y, z accelerometer
    float gx, gy, gz;		// x, y, z gyroscope
	float pitch = 0.0, roll = 0.0, yaw = 0.0; // Orientation angles
  float displacement[3] = {0.0, 0.0, 0.0};
  float G[3] = {0.0, 0.0, 9.81};
  float R_inv[3][3];

// Filtered value
float filtered = 0;


void setup() {
  Serial.begin(115200);
  BLE.begin();
  pinMode(LED_BUILTIN, OUTPUT);

  if (!IMU.begin()) {
    Serial.println("Failed to initializ_ae IMU!");
    while (1);
  }
  Serial.println("IMU initializ_aed.");

  // Set the BLE device name
  BLE.setLocalName(BLE_LOCAL_NAME);
  BLE.setDeviceName(BLE_DEVICE_NAME);

  // Add the service and characteristic
  BLE.setAdvertisedService(Service);
  Service.addCharacteristic(chaAx);
  Service.addCharacteristic(chaAy);
  Service.addCharacteristic(chaAz);
  Service.addCharacteristic(chaGx);
  Service.addCharacteristic(chaGy);
  Service.addCharacteristic(chaGz);

  BLE.addService(Service);

  chaAx.writeValue(0);
  chaAy.writeValue(0);
  chaAz.writeValue(0);
  chaGx.writeValue(0);
  chaGy.writeValue(0);
  chaGz.writeValue(0);

    // Start advertising the service
  BLE.advertise();
  Serial.println("Advertising started");

}

void loop() {

  static unsigned long lastSampleTime = 0;
  if (millis() - lastSampleTime >= sampling) {
    lastSampleTime = millis();
    // Maintain BLE connections
    BLEDevice central = BLE.central();
    central.connected();
    BLE.poll();
    
    if (IMU.accelerationAvailable()) {
      IMU.readAcceleration(ax, ay, az);
	    float acc_vector[3] = {ax, ay, az};
	  
	  
      // Calculate pitch and roll from accelerometer
      pitch = atan2(-ax, sqrt(ay * ay + az * az));
      roll = atan2(-ay, sqrt(ax * ax + az * az));
	
      IMU.readGyroscope(gx, gy, gz);
	  
	    gx = radians(gx);
		  gy = radians(gy);
		  gz = radians(gz);
		
      yaw += gz * sampling; // Adjust yaw over time using gyro Z data

      // Rotation Matrix
      float R[3][3] = {
      {cos(yaw) * cos(pitch), 	cos(yaw) * sin(pitch) * sin(roll) - sin(yaw) * cos(roll), 	cos(yaw) * sin(pitch) * cos(roll) + sin(yaw) * sin(roll)},
      {sin(yaw) * cos(pitch), 	sin(yaw) * sin(pitch) * sin(roll) + cos(yaw) * cos(roll), 	sin(yaw) * sin(pitch) * cos(roll) - cos(yaw) * sin(roll)},
      {-sin(pitch)		  ,						cos(pitch) * sin(roll)					,					cos(pitch) * cos(roll)						}
      };
      // From https://web.archive.org/web/20180112063505/http://www.chrobotics.com/library/accel-position-velocity
      // to compensate the gravity I must compute displacement = R_inv*acc_vector + (0,0,g)

      //Invert R -> 1/R = transpose(R) since is orthogonal
        for (int i = 0; i < 3; i++) {
          for (int j = 0; j < 3; j++) {
            R_inv[i][j] = R[j][i];
          }
        }


      for (int i = 0; i < 3; i++) {
        displacement[i] = 0;
        for (int j = 0; j < 3; j++) {
          displacement[i] += (R_inv[i][j] * acc_vector[j]);
        }
	    }

        for (int i = 0; i < 3; i++) {
          displacement[i] += G[i];
        }

      float abs = sqrt(displacement[1]*displacement[1] + displacement[2]*displacement[2] + displacement[3]*displacement[3]);
      filtered = butterworthBandpassFilter(abs);

      acc_Buffer[bufferIndex_a] = filtered;
      
      bufferIndex_a = (bufferIndex_a + 1) % (window / sampling);

      // Print counting and call count function
      if (bufferIndex_a == 0) {
        int n_movements = counting(acc_Buffer, window / sampling);
        Serial.print("# movements per minute: ");
        Serial.println(n_movements);

        Serial.print(" Abs: ");
        Serial.println(abs);

        Serial.print(" x_a: ");
        Serial.println(displacement[1]);
        
        Serial.print(" y_a: ");
        Serial.println(displacement[2]);

        Serial.print(" z_a: ");
        Serial.println(displacement[3]);

        digitalWrite(LED_BUILTIN, HIGH);  // turn the LED on (HIGH is the voltage level)
        delay(100);                      // wait for a second
        digitalWrite(LED_BUILTIN, LOW);   // turn the LED off by_a making the voltage LOW
        delay(100);                      // wait for a second

      chaAx.writeValue(ax);
      chaAy.writeValue(ay);
      chaAz.writeValue(az); 
      chaGx.writeValue(gx);
      chaGy.writeValue(gy);
      chaGz.writeValue(gz); 
      }
    }
  }
}

// Function to apply the Butterworth bandpass filter (10th-order)
float butterworthBandpassFilter(float input) {
  // Shift previous inputs and outputs
  for (int i = 10; i > 0; i--) {
    x[i] = x[i-1];
    y[i] = y[i-1];
  }

  // Set the new input value
  x[0] = input;

  // Apply the filter equation: y[n] = sum(b[i]*x[n-i]) - sum(a[i]*y[n-i])
  float output = 0;
  for (int i = 0; i < 11; i++) {
    output += b[i] * x[i];
    if (i > 0) {
      output -= a[i] * y[i];
    }
  }

  // Store the output value
  y[0] = output;

  return output;
}


// Function to calculate displacement rate from acceleration data
int counting(float *buffer, int size) {
  int zeroCrossings = 0;
  for (int i = 1; i < size; i++) {
    if ((buffer[i - 1] < threshold && buffer[i] > threshold) || (buffer[i - 1] > threshold && buffer[i] < threshold)) {
      zeroCrossings++;
    }
  }

  // Convert zero crossings to movements per minute
  int movementsPerMinute = (zeroCrossings / 2) * (60000 / window);
  return movementsPerMinute;
}
